﻿// Dependencies
//#region Dependencies

// External (Common) dependencies
var moment = require('moment');
var fs = require('fs');
var cp = require('child_process');

// External (Device) dependencies
var serialport = require("serialport");
var ieee754 = require('ieee754');
var Parser = require('expr-eval').Parser;// Library to manage parsing and evaluation of formulas
var mathparser = new Parser();

// Globals
var SummitInfo;// Summit Info (Name, MACID)
var LogManager;// Used for logging
var SocketIoManager;// Used for messaging
var SmtpManager;// Used for SMTP messaging
var SnmpManager;// Used for SNMP messaging - base library
//#endregion Dependencies

// Constants
//#region Constants

// Constants (Common)
var DEFAULT_DEVICE_MODEL_REQUIRE_PATH = "./DirectionalPower+VSWR.json";//  Device model defines the device incl. general properties, parameters and I/O
var PERSIST_DEVICE_MODEL_REQUIRE_PATH = "../data/port";// Relative to this module, so note the up directory
var PERSIST_DEVICE_MODEL_FS_PATH = "./data/port";// Relative to node, so note current directory
var DELAY_DEVICEINFO = 30000;//2000 for testing, 30000 for real-world - give time for the device to power up and be ready
var DELAY_PROCESSING = 45000;//3000 for testing, 45000 for real-world - to give time for device info

//#region for testing
var xxxForwardPower = 1;
var xxxForwardPowerReset = 0;
var xxxForwardPowerHigh = 10;
var xxxForwardPowerIncrement = 0.25;
var xxxReflectedPower = 1;
var xxxReflectedPowerReset = 0;
var xxxReflectedPowerHigh = 4;
var xxxReflectedPowerIncrement = 0.25;
var xxxVSWR = 1;
var xxxVSWRReset = 1;
var xxxVSWRHigh = 1.22;
var xxxVSWRIncrement = 0.02;
var xxxTemperature = -5;
var xxxTemperatureReset = -5;
var xxxTemperatureHigh = 35;
var xxxTemperatureIncrement = 1;
//#endregion for testing

var FORWARD_POWER_FIRST_READ = true;
var REFLECTED_POWER_FIRST_READ = true;
var VSWR_FIRST_READ = true;
var TEMPERATURE_FIRST_READ = true;

var VSWR_GOOD = 0
var VSWR_ERROR_REFLECTED_GREATER_FORWARD = 1
var VSWR_ERROR_FORWARD_TOO_LOW = 2
var VSWR_ERROR_VSWR_INFINITY = 3
var VSWR_ERROR_VSWR_LESS_THAN_1 = 4

var SERIAL_PORT_LOAD_DEFAULT_MODEL = 0;
var SERIAL_PORT_LOAD_PERSISTED_MODEL = 1;

var DEVICE_IN_RANGE = 0;
var DEVICE_BELOW_THRESHOLD = 1;
var DEVICE_ABOVE_THRESHOLD = 2;
var DEVICE_MINOR_THRESHOLD = 3;
var DEVICE_MAJOR_THRESHOLD = 4;
var DEVICE_UNKNOWN_THRESHOLD = 5;
var ThresholdValues = ["NORMAL", "BELOW", "ABOVE", "MINOR", "MAJOR","UNKNOWN"];

// Protocol Specific Constants
//#region Protocol Specific Constants
// Protocol Addresses
var DEVICE_ADDRESS_BROADCAST        = 254;// 254 is the Directional Power + VSWR device address for a broadcast command
var DEVICE_ADDRESS_DEFAULT          = 255;// 255 is the Directional Power + VSWR device default address

// Protocol Command
DEVICE_COMMAND_SET_BUS_ADDRESS      = 1;
DEVICE_COMMAND_GET_SERIAL_NUMBER    = 2;
DEVICE_COMMAND_GET_NAME             = 4;
DEVICE_COMMAND_GET_FORWARD_POWER    = 10
DEVICE_COMMAND_GET_REFLECTED_POWER  = 11
DEVICE_COMMAND_GET_VSWR             = 12
DEVICE_COMMAND_GET_MODEL            = 16
DEVICE_COMMAND_GET_TEMPERATURE      = 29
DEVICE_COMMAND_ZERO_OFFSETS         = 37
DEVICE_COMMAND_GET_FIRMWARE_VERSION = 43;

//#endregion Protocol Specific Constants

var EVENTS_LENGTH = 10;


// SNMP - this is here for now, could move
var AuthProtocol = ["None", "MD5", "SHA"];
var PrivProtocol = ["None", "DES", "AES"];

//#endregion Constants

// Device Class: Variables, Instantiation, Methods and Constructor
//#region Device Class: Variables, Instantiation, Methods and Constructor
function Device(port, loadFile, logger, io, info, smtp, snmp) {

    // Class Variables
    //#region Class Variables

    // Class Variables (Common)
    //#region Class Variables (Common)
    var EventLog = [];
    var EnabledFlag = false;
    var SendEventsFlag = false;
    var DeviceModel;
    var ThisDeviceObject = this;
    var VSWRStatus = VSWR_GOOD;
    //#endregion Class Variables (Common)

    //#endregion Class Variables

    // Class Instantiation
    //#region Class Instantiation

    // Local Dependencies
    //#region Local Dependencies
    LogManager = logger;
    SocketIoManager = io;
    SummitInfo = info;
    SmtpManager = smtp;
    SnmpManager = snmp;
    //#endregion Local Dependencies

    try {
        // Load a private model for this instance of the device. Two cases:
        //      1. New model: load from default and save for future use
        //      2. Existing model: load from persisted location

        var deviceModelFileLocation;
        if (loadFile == SERIAL_PORT_LOAD_DEFAULT_MODEL) {
            deviceModelFileLocation = DEFAULT_DEVICE_MODEL_REQUIRE_PATH;
        }
        else if (loadFile == SERIAL_PORT_LOAD_PERSISTED_MODEL) {
            deviceModelFileLocation = PERSIST_DEVICE_MODEL_REQUIRE_PATH + port + ".json";
        }

        try {
            //  Remove model from cache to avoid errors
            var cache = require.resolve(deviceModelFileLocation);
            if (cache) delete require.cache[cache];
            LogManager.trace("Successfully cleared device cache for port=" + port);
        }
        catch (error) {
            LogManager.error("Failed clearing device cache for port=" + port + " error=" + error);
        }

        // Load model, as set above 1. default or 2. persisted
        var DeviceModel = require(deviceModelFileLocation);

        //  Assign statics to model
        DeviceModel.port = port;
        DeviceModel.uuid = parseInt(moment().format("x"));

        // Restart hold timers
        DeviceModel.parameters.ForwardPower.HoldTimerActive = false;
        DeviceModel.parameters.ReflectedPower.HoldTimerActive = false;
        DeviceModel.parameters.VSWR.HoldTimerActive = false;
        DeviceModel.parameters.Temperature.HoldTimerActive = false;

        //  If we are loading from the default model,
        //  save a persisted model for the port
        if (loadFile == SERIAL_PORT_LOAD_DEFAULT_MODEL) {
            // Put creation time in TimeChanged variables, will later be updated on value change 
            var timeChanged = moment();
            DeviceModel.parameters.ForwardPower.TimeChanged = timeChanged;
            DeviceModel.parameters.ReflectedPower.TimeChanged = timeChanged;
            DeviceModel.parameters.VSWR.TimeChanged = timeChanged;
            DeviceModel.parameters.Temperature.TimeChanged = timeChanged;

            // Clear and then save file, sync to ensure persistence
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor ('" + deviceModelFileLocation + "') - before saveDeviceModel");
            fs.truncate(PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json", 0, function () {
                fs.writeFile(PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json", JSON.stringify(DeviceModel), function (error) {
                    if (error) {
                        LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: write error '" + PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json' " + error);
                    }
                    cp.exec("sync", function (error, stdout, stderr) {
                        if (error) {
                            LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: sync error '" + PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json' " + error);
                        } else {
                            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor: Device Updated! '" + PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json'");
                        }
                    });
                });
            });
        }

        //  Log constructor activity
        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor ('" + deviceModelFileLocation + "'): Device uuid=" + DeviceModel.uuid + " ------------------------------------------------------------");
        LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Constructor ('" + deviceModelFileLocation + "'): DeviceModel=" + JSON.stringify(DeviceModel));
    }
    catch (error) {
        LogManager.error("Failed constructing device on port=" + port + " error=" + error);
    }
    //#endregion Class Instantiation

    // Class Methods
    //#region Class Methods

    //#region Class Methods (Common)

    //#region Class Methods (Common - Basic)
    this.setPort = function (port) {
        DeviceModel.port = port;
    }

    this.getPort = function () {
        return DeviceModel.port;
    }

    this.getDeviceId = function () {
        LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".getDeviceId: Device id=" + DeviceModel.id);
        return DeviceModel.id;
    }

    this.getDeviceModel = function () {
        LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".getDeviceModel: Device uuid=" + DeviceModel.uuid);
        return DeviceModel;
    }

    this.enable = function (enabled) {
        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".Enable: Enabled=" + (enabled ? "TRUE" : "FALSE"));
        if (enabled == EnabledFlag)
            return;// Don't do anything if we're already in the proper state
        else
            EnabledFlag = enabled;// Remember new setting

        if (enabled) {
            //#region Enable Serial Port
            if (!DirectionalPowerVswrSerialPort.isOpen()) {
                DirectionalPowerVswrSerialPort.open(function (error) {
                    if (error) {
                        LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[true]: Serial Port open error=" + error);
                    } else {
                        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[true]: Serial Port opened");

                        // Get device information
                        setTimeout(function () { ThisDeviceObject.getDeviceInformationFromDevice() }, DELAY_DEVICEINFO);// Give this device 30 seconds to bootup

                        // Start processing 5seconds later after we get device information
                        setTimeout(function () { ThisDeviceObject.startProcessing() }, DELAY_PROCESSING);// Wait another 10 seconds until we've got the Device Info
                    }
                });
            } else {
                // Serial port is already open
                LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[true]: Serial Port is already open.");

                // Get device information
                setTimeout(function () { ThisDeviceObject.getDeviceInformationFromDevice() }, DELAY_DEVICEINFO);// Give this device 30 seconds to bootup

                // Start processing 5seconds later after we get device information
                setTimeout(function () { ThisDeviceObject.startProcessing() }, DELAY_PROCESSING);// Wait another 10 seconds until we've got the Device Info
            }
            //#endregion Enable Serial Port
        }

        else {
            this.stopProcessing();// Stop serial port commands

            //#region Disable Serial Port
            if (DirectionalPowerVswrSerialPort.isOpen()) {
                DirectionalPowerVswrSerialPort.close(function (error) {
                    if (error) {
                        LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[false]: Serial Port close error=" + error);
                    } else {
                        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".enable[false]: Serial Port closed!");
                    }
                });
            }
            //#endregion Disable Serial Port
        }
    }

    this.reset = function (callback) {
        // Pause processing
        PauseDeviceInfoControl = true;
        PauseParametersCycleControl = true;

        // Issue a reset command
        DirectionalPowerVswrSendParamsBuffer[0] = DEVICE_ADDRESS_BROADCAST;// Broadcast address
        DirectionalPowerVswrSendParamsBuffer[1] = DEVICE_COMMAND_SET_BUS_ADDRESS;// Set buss address command
        DirectionalPowerVswrSendParamsBuffer[2] = 1;// One byte
        DirectionalPowerVswrSendParamsBuffer[3] = DEVICE_ADDRESS_DEFAULT;// Default address

        crc = this.CalculateCRC(DirectionalPowerVswrSendParamsBuffer, 4);
        DirectionalPowerVswrSendParamsBuffer[4] = (crc >> 8) & 0xFF;
        DirectionalPowerVswrSendParamsBuffer[5] = crc & 0xFF;

        ExpectedResponseLength = 6;// Expect to get 6 bytes back from device for this command

        DirectionalPowerVswrSerialPort.write(DirectionalPowerVswrSendParamsBuffer, function (error) {
            if (error) {
                LogManager.error("this.reset: Error writing to serial port - Error=" + error);
                ThisDeviceObject.RecoverUselessSerialPort();
                //for (i = 0; i < 5; i++) {
                //    Logger.error("this.DirectionalPowerVswrGetParameters: Sending Buffer[" + i + "]=" + DirectionalPowerVswrSendBasicBuffer[i]);
                //}
                return;
            }
            LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".reset: Reset command sent to device: " + DeviceModel.id + " on Port=" + Number(DeviceModel.port+1));

            // Assume everything completed in 1 second
            setTimeout(function () {
                var status = 0;

                // Restart processing
                PauseDeviceInfoControl = false;
                PauseParametersCycleControl = false;
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".reset: Restarting device processing for device: " + DeviceModel.id + " on Port=" + Number(DeviceModel.port + 1));

                callback(status);// Default to return success for now
            }, 1000);
        });
    }

    this.zero = function (callback) {
        // Pause processing
        PauseDeviceInfoControl = true;
        PauseParametersCycleControl = true;

        // Issue a reset command
        DirectionalPowerVswrSendParamsBuffer[0] = DEVICE_ADDRESS_BROADCAST;// Broadcast address
        DirectionalPowerVswrSendParamsBuffer[1] = DEVICE_COMMAND_ZERO_OFFSETS;// Zero the offsets
        DirectionalPowerVswrSendParamsBuffer[2] = 0;// No data
        crc = this.CalculateCRC(DirectionalPowerVswrSendParamsBuffer, 3);
        DirectionalPowerVswrSendParamsBuffer[4] = (crc >> 8) & 0xFF;
        DirectionalPowerVswrSendParamsBuffer[5] = crc & 0xFF;

        ExpectedResponseLength = 0;// Doesn't appear that this command returns anything

        DirectionalPowerVswrSerialPort.write(DirectionalPowerVswrSendParamsBuffer, function (error) {
            if (error) {
                LogManager.error("this.zero: Error writing to serial port - Error=" + error);
                ThisDeviceObject.RecoverUselessSerialPort();
                return;
            }
            LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".zero: Zero offsets command sent to device: " + DeviceModel.id + " on Port=" + Number(DeviceModel.port+1));

            // Assume everything completed in 1 second
            setTimeout(function () {
                var status = 0;

                // Restart processing
                PauseDeviceInfoControl = false;
                PauseParametersCycleControl = false;
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".zero: Restarting device processing for device: " + DeviceModel.id + " on Port=" + Number(DeviceModel.port+1));

                callback(status);// Default to return success for now
            }, 1000);
        });
    }

    this.sendEvents = function (sendevents) {
        SendEventsFlag = sendevents;
        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendEvents: SendEventsFlag=" + SendEventsFlag);
    }
    //#endregion Class Methods (Common - Basic)

    //#region Class Methods (Common - UpdateDeviceModel)
    this.updateDeviceModel = function (port, model, callback) {

        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateDeviceModel: Device uuid=" + DeviceModel.uuid);

        // If the parameter has been enabled, re-evaluate the threshold
        if (model.parameters.ForwardPower.Enabled && !DeviceModel.parameters.ForwardPower.Enabled) {
            model.parameters.ForwardPower.ThresholdState = DEVICE_UNKNOWN_THRESHOLD;
        }
        if (model.parameters.ReflectedPower.Enabled && !DeviceModel.parameters.ReflectedPower.Enabled) {
            model.parameters.ReflectedPower.ThresholdState = DEVICE_UNKNOWN_THRESHOLD;
        }
        if (model.parameters.VSWR.Enabled && !DeviceModel.parameters.VSWR.Enabled) {
            model.parameters.VSWR.ThresholdState = DEVICE_UNKNOWN_THRESHOLD;
        }
        if (model.parameters.Temperature.Enabled && !DeviceModel.parameters.Temperature.Enabled) {
            model.parameters.Temperature.ThresholdState = DEVICE_UNKNOWN_THRESHOLD;
        }

        // Update the device model and recalcualte threshold values
        DeviceModel = model;
        ThisDeviceObject.recalculateFormulaThresholdValues();

        // Restart hold timers
        DeviceModel.parameters.ForwardPower.HoldTimerActive = false;
        DeviceModel.parameters.ReflectedPower.HoldTimerActive = false;
        DeviceModel.parameters.VSWR.HoldTimerActive = false;
        DeviceModel.parameters.Temperature.HoldTimerActive = false;

        // Re-assess values based on potential threshold changes
        ThisDeviceObject.updateForwardPower(DeviceModel.parameters.ForwardPower.RawValue);
        ThisDeviceObject.updateReflectedPower(DeviceModel.parameters.ReflectedPower.RawValue);
        ThisDeviceObject.updateVSWR(DeviceModel.parameters.VSWR.RawValue);
        ThisDeviceObject.updateTemperature(DeviceModel.parameters.Temperature.RawValue);

        // Persist the new DeviceModel
        ThisDeviceObject.saveDeviceModel(DeviceModel);
        callback(DeviceModel);
    }

    this.recalculateFormulaThresholdValues = function () {
        var _expression;// Formula expression value 
        var i;

        // ForwardPower
        DeviceModel.parameters.ForwardPower.Formulas[0].FormulaLowThreshold = Number(DeviceModel.parameters.ForwardPower.Formulas[0].FormulaLowThreshold).toFixed(DeviceModel.parameters.ForwardPower.Formulas[0].Decimals);
        DeviceModel.parameters.ForwardPower.Formulas[0].FormulaHighThreshold = Number(DeviceModel.parameters.ForwardPower.Formulas[0].FormulaHighThreshold).toFixed(DeviceModel.parameters.ForwardPower.Formulas[0].Decimals);
        for (i = 1; i < DeviceModel.parameters.ForwardPower.Formulas.length; i++) {
            _expression = mathparser.parse(DeviceModel.parameters.ForwardPower.Formulas[i].Formula);
            DeviceModel.parameters.ForwardPower.Formulas[i].FormulaLowThreshold = Number(_expression.evaluate({ x: DeviceModel.parameters.ForwardPower.Formulas[0].FormulaLowThreshold })).toFixed(DeviceModel.parameters.ForwardPower.Formulas[i].Decimals);
            DeviceModel.parameters.ForwardPower.Formulas[i].FormulaHighThreshold = Number(_expression.evaluate({ x: DeviceModel.parameters.ForwardPower.Formulas[0].FormulaHighThreshold })).toFixed(DeviceModel.parameters.ForwardPower.Formulas[i].Decimals);
        }

        // ReflectedPower
        DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaLowThreshold = Number(DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaLowThreshold).toFixed(DeviceModel.parameters.ReflectedPower.Formulas[0].Decimals);
        DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaHighThreshold = Number(DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaHighThreshold).toFixed(DeviceModel.parameters.ReflectedPower.Formulas[0].Decimals);
        for (i = 1; i < DeviceModel.parameters.ReflectedPower.Formulas.length; i++) {
            _expression = mathparser.parse(DeviceModel.parameters.ReflectedPower.Formulas[i].Formula);
            DeviceModel.parameters.ReflectedPower.Formulas[i].FormulaLowThreshold = Number(_expression.evaluate({ x: DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaLowThreshold })).toFixed(DeviceModel.parameters.ReflectedPower.Formulas[i].Decimals);
            DeviceModel.parameters.ReflectedPower.Formulas[i].FormulaHighThreshold = Number(_expression.evaluate({ x: DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaHighThreshold })).toFixed(DeviceModel.parameters.ReflectedPower.Formulas[i].Decimals);
        }

        // VSWR
        DeviceModel.parameters.VSWR.Formulas[0].FormulaLowThreshold = Number(DeviceModel.parameters.VSWR.Formulas[0].FormulaLowThreshold).toFixed(DeviceModel.parameters.VSWR.Formulas[0].Decimals);
        DeviceModel.parameters.VSWR.Formulas[0].FormulaHighThreshold = Number(DeviceModel.parameters.VSWR.Formulas[0].FormulaHighThreshold).toFixed(DeviceModel.parameters.VSWR.Formulas[0].Decimals);

        // Temperature
        DeviceModel.parameters.Temperature.Formulas[0].FormulaLowThreshold = Number(DeviceModel.parameters.Temperature.Formulas[0].FormulaLowThreshold).toFixed(DeviceModel.parameters.Temperature.Formulas[0].Decimals);
        DeviceModel.parameters.Temperature.Formulas[0].FormulaHighThreshold = Number(DeviceModel.parameters.Temperature.Formulas[0].FormulaHighThreshold).toFixed(DeviceModel.parameters.Temperature.Formulas[0].Decimals);
        for (i = 1; i < DeviceModel.parameters.Temperature.Formulas.length; i++) {
            _expression = mathparser.parse(DeviceModel.parameters.Temperature.Formulas[i].Formula);
            DeviceModel.parameters.Temperature.Formulas[i].FormulaLowThreshold = Number(_expression.evaluate({ x: DeviceModel.parameters.Temperature.Formulas[0].FormulaLowThreshold })).toFixed(DeviceModel.parameters.Temperature.Formulas[i].Decimals);
            DeviceModel.parameters.Temperature.Formulas[i].FormulaHighThreshold = Number(_expression.evaluate({ x: DeviceModel.parameters.Temperature.Formulas[0].FormulaHighThreshold })).toFixed(DeviceModel.parameters.Temperature.Formulas[i].Decimals);
        }

    }
    //#endregion Class Methods (Common - UpdateDeviceModel)

    //#endregion Class Methods (Common)

    //#region Class Methods (Device Processing)

    //#region Device Processing Constants

    // Serial Port Control
    //#region Serial Port Control
    var ExpectedResponseLength = 0;
    var CurrentResponseBufferPtr = 0;
    var ResponseBuffer = new Buffer(256);
    var DirectionalPowerVswrSendBasicBuffer = new Uint8Array(5);
    var DirectionalPowerVswrSendParamsBuffer = new Uint8Array(6);
    var DirectionalPowerVswrSerialPort = new serialport("/dev/ttyUSB" + DeviceModel.port, { baudRate: 57600, dataBits: 8, stopBits: 1, parity: 'none', parser: serialport.parsers.raw, autoOpen: false });
    //#endregion Serial Port Control

    // Processing Cycle (setInterval) Control
    //#region Processing Cycle (setInterval) Control
    var DeviceInfoControl;
    var PauseDeviceInfoControl = false;
    var GetParametersCycleControl;
    var PauseParametersCycleControl = false;
    //#endregion Processing Cycle (setInterval) Control

    // Device Information Processing
    //#region Device Information Processing
    var DirectionalPowerVswrDeviceInfoCommands = [DEVICE_COMMAND_GET_SERIAL_NUMBER, DEVICE_COMMAND_GET_MODEL, DEVICE_COMMAND_GET_FIRMWARE_VERSION,DEVICE_COMMAND_GET_NAME]
    var DirectionalPowerVswrDeviceInfoCommandsString = ["Serial Number", "Device Name", "Model Number", "Firmware Version"];
    var DirectionalPowerVswrDeviceInfoReturnLength = [9, 23, 7, 12]
    var DirectionalPowerVswrDeviceInfoCommandsIndex = 0;
    //#endregion Device Information Processing

    // Parameter Processing
    //#region Parameter Processing
    var DirectionalPowerVswrParameterCommands = [DEVICE_COMMAND_GET_FORWARD_POWER, DEVICE_COMMAND_GET_REFLECTED_POWER, DEVICE_COMMAND_GET_VSWR, DEVICE_COMMAND_GET_TEMPERATURE]
    var DirectionalPowerVswrParameterCommandsString = ["Forward Power", "Reflected Power", "VSWR", "Temperature"];
    var DirectionalPowerVswrParameterReturnLength = [9, 9, 9, 9]
    var DirectionalPowerVswrParameterCommandsIndex = 0;
    //#endregion Parameter Processing

    //#endregion Device Processing Constants

    //#region Class Methods (Device Processing - Start/Stop)
    //#region startProcessing
    this.getDeviceInformationFromDevice = function () {
        // Serial Cycle
        DeviceInfoControl = setInterval(function () {
            if (!PauseDeviceInfoControl) {// This is used to pause the cycle temporarily
                ThisDeviceObject.DirectionalPowerVswrGetDeviceInfo()
            }
        }, DeviceModel.parameters.DeviceInformation.CycleTime);
        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".getDeviceInformationFromDevice: Get Device Info - cycle=" + DeviceModel.parameters.DeviceInformation.CycleTime + "ms");
    }

    this.startProcessing = function () {
        //  Start Serial IO - Get Parameter loop
        //#region Start Serial IO - Get Parameter loop
        GetParametersCycleControl = setInterval(function () {
            if (!PauseParametersCycleControl) {// This is used to pause the cycle temporarily
                ThisDeviceObject.DirectionalPowerVswrGetParameters()
            }
        }, DeviceModel.parameters.ForwardPower.CycleTime);
        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".startProcessing: IO processing cycle=" + DeviceModel.parameters.ForwardPower.CycleTime + "ms");
        //#endregion Start Serial IO - Get Parameter loop
    }
    //#endregion startProcessing

    //#region stopProcessing
    this.stopProcessing = function () {

        //  Stop serial IO processing loops
        //#region  Stop serial IO processing loops
        try {
            clearInterval(DeviceInfoControl);
            LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopping get Device Information cycle.");
            clearInterval(GetParametersCycleControl);
            LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopping IO processing cycle.");
        } catch (error) {
            LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Error stopping DeviceInfo and IO processing cycle. Error=" + error);
        } finally {
            LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".stopProcessing: Stopped DeviceInfo and IO processing cycle");
        }
        //#endregion  Stop serial IO processing loops
    }
    //#endregion stopProcessing
    //#endregion Class Methods (Device Processing - Start/Stop)

    //#region Class Methods (Device Processing - Device Information)
    //#region getDeviceInformation


    this.getDeviceInformation = function () {
        return DeviceModel.parameters.DeviceInformation;
    }
    //#endregion getDeviceInformation

    //#endregion Class Methods (Device Processing - Device Information)

    //#region Class Methods (Device Processing - Process I/O)

    //#region Serial Port Communications
    DirectionalPowerVswrSerialPort.on('open', function () {
        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort: Opened");
    });

    // Workaround for Mantis #993 - Write Errors seen after continuous operation
    this.RecoverUselessSerialPort = function () {
        LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".RecoverUselessSerialPort: Failed write problem detected and workaround started");
        // Pause polling operation, while we recover the serial port
        PauseDeviceInfoControl = true;
        PauseParametersCycleControl = true;

        // Close the serial port
        try {
            // Probably needs to do an isOpen check in the future ...
            DirectionalPowerVswrSerialPort.close();
        } catch (error) {
            LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".RecoverUselessSerialPort: Serial Port close error=" + error);
        }

        // Re-open the serial port
        DirectionalPowerVswrSerialPort.open(function (error) {
            if (error) {
                LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".RecoverUselessSerialPort: Serial Port open error=" + error);
            } else {
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".RecoverUselessSerialPort: Serial Port re-opened");
            }
        });

        // Re-enable polling operation by un-pausing polling operation
        PauseDeviceInfoControl = false;
        PauseParametersCycleControl = false;
    }

    this.DirectionalPowerVswrGetDeviceInfo = function () {
        // Check to see if we've been disabled
        if (!EnabledFlag) {
            try {
                clearInterval(DeviceInfoControl);
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrGetParameters: Stopping GetDeviceInfo (DeviceInfoControl)");
            }
            catch (error) {
                LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrGetParameters: Error stopping GetDeviceInfo (DeviceInfoControl) error=" + error);
            }
            finally {
                LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrGetParameters: Stopped GetDeviceInfo (DeviceInfoControl)");
                return;
            }
        }

        // Send next command in the cycle
        //#region Send next command in the cycle
        DirectionalPowerVswrSendBasicBuffer[0] = DeviceModel.parameters.DeviceInformation.DeviceAddress;  // Device Addrsess - for this release we'll use default value in JSON file (probably 255)
        DirectionalPowerVswrSendBasicBuffer[1] = DirectionalPowerVswrDeviceInfoCommands[DirectionalPowerVswrDeviceInfoCommandsIndex];
        DirectionalPowerVswrSendBasicBuffer[2] = 0;
        crc = this.CalculateCRC(DirectionalPowerVswrSendBasicBuffer, 3);
        DirectionalPowerVswrSendBasicBuffer[3] = (crc >> 8) & 0xFF;
        DirectionalPowerVswrSendBasicBuffer[4] = crc & 0xFF;

        // Set how much data we expect to receive
        ExpectedResponseLength = DirectionalPowerVswrDeviceInfoReturnLength[DirectionalPowerVswrDeviceInfoCommandsIndex];
        //Logger.info("this.DirectionalPowerVswrGetDeviceInfo: Command=" + DirectionalPowerVswrDeviceInfoCommands[DirectionalPowerVswrDeviceInfoCommandsIndex] + " (" + DirectionalPowerVswrDeviceInfoCommandsString[DirectionalPowerVswrDeviceInfoCommandsIndex] + ")");

        //for (var i = 0; i < 5; i++) {
        //    logger.info("this.DirectionalPowerVswrGetDeviceInfo: Sending Buffer[" + i + "]=" + DirectionalPowerVswrSendBasicBuffer[i]);
        //}

        // Stop processing once we're done the command in the list
        if (++DirectionalPowerVswrDeviceInfoCommandsIndex == DirectionalPowerVswrDeviceInfoCommands.length) clearInterval(DeviceInfoControl);

        DirectionalPowerVswrSerialPort.write(DirectionalPowerVswrSendBasicBuffer, function (error) {
            if (error) {
                logger.error("this.DirectionalPowerVswrGetDeviceInfo: Error writing to serial port - Error=" + error);
                ThisDeviceObject.RecoverUselessSerialPort();
                //for (i = 0; i < 5; i++) {
                //    Logger.error("this.DirectionalPowerVswrGetParameters: Sending Buffer[" + i + "]=" + DirectionalPowerVswrSendBasicBuffer[i]);
                //}
                return;
            }
        });
        //#endregion Send next command in the cycle
    }

    this.DirectionalPowerVswrGetParameters = function () {
        // Check to see if we've been disabled
        if (!EnabledFlag) {
            try {
                clearInterval(GetParametersCycleControl);
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrGetParameters: Stopping GetParameters (GetParametersCycleControl)");
            }
            catch (error) {
                LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrGetParameters: Error stopping GetParameters (GetParametersCycleControl) error=" + error);
            }
            finally {
                LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrGetParameters: Stopped GetParameters (GetParametersCycleControl)");
                return;
            }
        }

        // Send next command in the cycle
        //#region Send next command in the cycle
        DirectionalPowerVswrSendBasicBuffer[0] = DeviceModel.parameters.DeviceInformation.DeviceAddress;  // Device Address - for this release we'll use default value in JSON file (probably 255)
        DirectionalPowerVswrSendBasicBuffer[1] = DirectionalPowerVswrParameterCommands[DirectionalPowerVswrParameterCommandsIndex]; // Forward Power, Reflected Power, VSWR, Temperature ... then do it again
        DirectionalPowerVswrSendBasicBuffer[2] = 0;
        crc = this.CalculateCRC(DirectionalPowerVswrSendBasicBuffer, 3);
        DirectionalPowerVswrSendBasicBuffer[3] = (crc >> 8) & 0xFF;
        DirectionalPowerVswrSendBasicBuffer[4] = crc & 0xFF;

        // Set how much data we expect to receive
        ExpectedResponseLength = DirectionalPowerVswrParameterReturnLength[DirectionalPowerVswrParameterCommandsIndex];
        //Logger.info("this.DirectionalPowerVswrGetDeviceInfo: Command=" + DirectionalPowerVswrParameterCommands[DirectionalPowerVswrParameterCommandsIndex] + " (" + DirectionalPowerVswrParameterCommandsString[DirectionalPowerVswrParameterCommandsIndex] + ")");

        //for (i = 0; i < 5; i++) {
        //    logger.info("this.DirectionalPowerVswrGetParameters: Sending Buffer[" + i + "]=" + DirectionalPowerVswrSendBasicBuffer[i]);
        //}

        // Reset the index
        if (++DirectionalPowerVswrParameterCommandsIndex == DirectionalPowerVswrParameterCommands.length) DirectionalPowerVswrParameterCommandsIndex = 0;

        DirectionalPowerVswrSerialPort.write(DirectionalPowerVswrSendBasicBuffer, function (error) {
            if (error) {
                LogManager.error("this.DirectionalPowerVswrGetParameters: Error writing to serial port - Error=" + error);
                ThisDeviceObject.RecoverUselessSerialPort();
                //for (i = 0; i < 5; i++) {
                //    Logger.error("this.DirectionalPowerVswrGetParameters: Sending Buffer[" + i + "]=" + DirectionalPowerVswrSendBasicBuffer[i]);
                //}
                return;
            }
        });
        //#endregion Send next command in the cycle
    }

    DirectionalPowerVswrSerialPort.on('data', function (data) {
        var timeChanged = moment();
        var _buffer;// Temporary buffer
        var _ieee754Value;// IEEE754 calculated value
        var _4BytesValue;// 4-byte little endian value
        var _1ByteValue;// 1-byte value
        var BufferString;// Human view of buffer
        var _expression;// Formula expression value 

        // If we have full data then process, else append and wait for more
        //Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort.data: Received=" + data.toString('hex') + " (Rx=" + data.length + ", Expected=" + ExpectedResponseLength + "), CurrentPtr=" + CurrentResponseBufferPtr);
        data.copy(ResponseBuffer, CurrentResponseBufferPtr, 0, data.length);
        CurrentResponseBufferPtr += data.length;
        if (CurrentResponseBufferPtr < ExpectedResponseLength) {
            // Copy received buffer into response, and then wait for more data
            //Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort.data: Partial=" + data.toString('hex') + " (Rx=" + data.length + ", Expected=" + ExpectedResponseLength + ")");

            // Check for adjusted Expected response length (sometimes strings can be smaller than maximum size)
            if (CurrentResponseBufferPtr > 2) {
                if (ResponseBuffer[2] != (ExpectedResponseLength - 5)) {
                    ExpectedResponseLength = ResponseBuffer[2] + 5;
                    //Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort.data: Adjusted ExpectedLength=" + ExpectedResponseLength);
                }
            }
            return;
        } else {
            // We've read the full length so go back to beginning of buffer
            //  *** and importantly, we can drop to the switch statement below to parse the data ***
            //Logger.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort.data: Full=" + ResponseBuffer.toString('hex') + " (Rx=" + CurrentResponseBufferPtr + ", Expected=" + ExpectedResponseLength + ")");
            CurrentResponseBufferPtr = 0;
        }

        // Parse the response
        switch (ResponseBuffer[1]) {
            case 1:// Set bus address
                _1ByteValue = ResponseBuffer[3];
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort.ResponseBuffer: Reset for Device Address=" + _1ByteValue);
                break;
            case 2:// SerialNumber
                _4BytesValue = (ResponseBuffer[6] << 24) + (ResponseBuffer[5] << 16) + (ResponseBuffer[4] << 8) + (ResponseBuffer[3])
                DeviceModel.parameters.DeviceInformation.SerialNumber = _4BytesValue;
                BufferString = (ResponseBuffer.toString('hex')).substr(0, (ResponseBuffer[2] + 5) * 2);
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort.ResponseBuffer: SerialNumber=" + DeviceModel.parameters.DeviceInformation.SerialNumber + " (Buffer=" + BufferString + ")");
                ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "DeviceInformation", { "DeviceInformation": DeviceModel.parameters.DeviceInformation });
                ThisDeviceObject.saveDeviceModel(DeviceModel);
                break;
            case 4:// Name
                var _NameAsciiBuffer = new Buffer(ExpectedResponseLength - 5);
                ResponseBuffer.copy(_NameAsciiBuffer, 0, 3, (ExpectedResponseLength - 2));
                DeviceModel.parameters.DeviceInformation.DeviceName = _NameAsciiBuffer.toString('utf8');
                BufferString = (ResponseBuffer.toString('hex')).substr(0, (ResponseBuffer[2] + 5) * 2);
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort.ResponseBuffer: DeviceName=" + DeviceModel.parameters.DeviceInformation.DeviceName + " (Buffer=" + BufferString + ")");
                ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "DeviceInformation", { "DeviceInformation": DeviceModel.parameters.DeviceInformation });
                ThisDeviceObject.saveDeviceModel(DeviceModel);
                break;
            case 16:// Model
                var _ModelAsciiBuffer = new Buffer(ExpectedResponseLength - 5);
                ResponseBuffer.copy(_ModelAsciiBuffer, 0, 3, (ExpectedResponseLength - 2));
                DeviceModel.parameters.DeviceInformation.Model = _ModelAsciiBuffer.toString('utf8');
                BufferString = (ResponseBuffer.toString('hex')).substr(0, (ResponseBuffer[2] + 5) * 2);
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort.ResponseBuffer: Model=" + DeviceModel.parameters.DeviceInformation.Model + " (Buffer=" + BufferString + ")");
                ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "DeviceInformation", { "DeviceInformation": DeviceModel.parameters.DeviceInformation });
                ThisDeviceObject.saveDeviceModel(DeviceModel);
                break;
            case 43:// Firmware
                var _FirmwareAsciiBuffer = new Buffer(ExpectedResponseLength - 5);
                ResponseBuffer.copy(_FirmwareAsciiBuffer, 0, 3, (ExpectedResponseLength - 2));
                DeviceModel.parameters.DeviceInformation.Firmware = _FirmwareAsciiBuffer.toString('utf8');
                BufferString = (ResponseBuffer.toString('hex')).substr(0, (ResponseBuffer[2] + 5) * 2);
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".DirectionalPowerVswrSerialPort.ResponseBuffer: Firmware=" + DeviceModel.parameters.DeviceInformation.Firmware + " (Buffer=" + BufferString + ")");
                ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "DeviceInformation", { "DeviceInformation": DeviceModel.parameters.DeviceInformation });
                ThisDeviceObject.saveDeviceModel(DeviceModel);
                break;

            case 10:// Forward Power
                _buffer = new Buffer([ResponseBuffer[3], ResponseBuffer[4], ResponseBuffer[5], ResponseBuffer[6]]);
                _ieee754Value = ieee754.read(_buffer, 0, true, 23, 4);

                // If enabled, process forward power, also record UnqualifiedValue for VSWR filtering
                DeviceModel.parameters.ForwardPower.UnqualifiedValue = _ieee754Value;
                if (DeviceModel.parameters.ForwardPower.Enabled) {
                    ThisDeviceObject.updateForwardPower(_ieee754Value);
                }
                break;
            case 11:// Reflected Power
                _buffer = new Buffer([ResponseBuffer[3], ResponseBuffer[4], ResponseBuffer[5], ResponseBuffer[6]]);
                _ieee754Value = ieee754.read(_buffer, 0, true, 23, 4);

                // If enabled, process reflected power, also record UnqualifiedValue for VSWR filtering
                DeviceModel.parameters.ReflectedPower.UnqualifiedValue = _ieee754Value;
                if (DeviceModel.parameters.ReflectedPower.Enabled) {
                    ThisDeviceObject.updateReflectedPower(_ieee754Value);
                }
                break;
            case 12:// VSWR
                _buffer = new Buffer([ResponseBuffer[3], ResponseBuffer[4], ResponseBuffer[5], ResponseBuffer[6]]);
                _ieee754Value = ieee754.read(_buffer, 0, true, 23, 4);

                // If enabled, process VSWR
                if (DeviceModel.parameters.VSWR.Enabled) {
                    ThisDeviceObject.updateVSWR(_ieee754Value);
                };
                break;
            case 29:// Temperature
                _buffer = new Buffer([ResponseBuffer[3], ResponseBuffer[4], ResponseBuffer[5], ResponseBuffer[6]]);
                _ieee754Value = ieee754.read(_buffer, 0, true, 23, 4);

                // If enabled, process temperature
                if (DeviceModel.parameters.Temperature.Enabled) {
                    ThisDeviceObject.updateTemperature(_ieee754Value);
                }
                break;
        }
    });

    this.CalculateCRC = function (buffer, length) {
        var crc = 0xFFFF;

        //logger.info("CalculateCRC: buffer.length=" + length);

        for (var pos = 0; pos < length; pos++) {
            crc ^= buffer[pos];
            //logger.info("CalculateCRC: pos=" + pos+ " CRC="+crc);

            for (var i = 8; i != 0; i--) {
                if ((crc & 0x0001) != 0) {
                    crc >>= 1;
                    crc ^= 0xA001;
                }
                else
                    crc >>= 1;
            }
        }
        return crc;
    }
    //#endregion Serial Port Communications

    //#region Parameter Filtering

    this.updateForwardPower = function (readValue) {
        var timeChanged = moment();

        // Threshold Change logic
        //  - ThresholdChanged = true or false, depending if we've changed a threshold
        //#region Threshold Change Logic
        var ThresholdChanged = false;
        var HoldTimeChanged = 0;

        var PreviousRawValue = DeviceModel.parameters.ForwardPower.RawValue;
        var PreviousThreshold = DeviceModel.parameters.ForwardPower.ThresholdState;
        var NewThreshold;
        if (readValue < DeviceModel.parameters.ForwardPower.Formulas[0].FormulaLowThreshold) {
            NewThreshold = DEVICE_BELOW_THRESHOLD;
        }
        else if (readValue > DeviceModel.parameters.ForwardPower.Formulas[0].FormulaHighThreshold) {
            NewThreshold = DEVICE_ABOVE_THRESHOLD;
        } else {
            NewThreshold = DEVICE_IN_RANGE;
        }
        if (PreviousThreshold != NewThreshold) {
            ThresholdChanged = true;
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: Forward Power=" + DeviceModel.parameters.ForwardPower.Formulas[0].FormulaValue + " " + DeviceModel.parameters.ForwardPower.Formulas[0].Units + " Raw=" + DeviceModel.parameters.ForwardPower.RawValue);
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: Threshold change - new=" + ThresholdValues[NewThreshold] + " old=" + ThresholdValues[PreviousThreshold]);
        } else {
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: Forward Power=" + DeviceModel.parameters.ForwardPower.Formulas[0].FormulaValue + " " + DeviceModel.parameters.ForwardPower.Formulas[0].Units + " Raw=" + DeviceModel.parameters.ForwardPower.RawValue);
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: Threshold change - new=" + ThresholdValues[NewThreshold] + " old=" + ThresholdValues[PreviousThreshold]);
        }
        //#endregion Threshold Change Logic

        // Qualification logic
        //  - QualifierValid = true or false, depending if qualified by digital input
        //  - DeviceModel.parameters.ForwardPower.HoldTimerActive = true or false, depending on qualifications status
        //  - DeviceModel.parameters.ForwardPower.HoldTimer = holds time of when first qualified, with or without qualifier
        //#region Qualification logic
        var QualifierValid = false;

        if (DeviceModel.parameters.ForwardPower.Qualifier > 0) {
            // Qualifier USED logic
            //#region Qualifier USED logic

            // HoldTime Logic for Qualifier USED
            //#region HoldTime Logic for Qualifier USED
            if (DeviceModel.parameters.ForwardPower.QualifierValue > 0) {
                //#region HoldTime Logic for Qualifier USED & TRUE
                QualifierValid = true;
                LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [QUALIFER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");

                if (DeviceModel.parameters.ForwardPower.HoldTime > 0) {
                    // If there is a hold time ...
                    if (DeviceModel.parameters.ForwardPower.HoldTimerActive === false) {
                        // If the hold timer not previously running ...
                        if (ThresholdChanged) {
                            // And the threshold changed ... start the timer
                            DeviceModel.parameters.ForwardPower.HoldTimer = timeChanged;
                            DeviceModel.parameters.ForwardPower.HoldTimerActive = true;
                            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                            ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerHoldState", { "ForwardPowerHoldState": "HOLDING" });
                        } else {
                            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                            ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerHoldState", { "ForwardPowerHoldState": "OFF" });
                        }
                    }
                    else {
                        // If the hold timer was already running ...
                        if (!ThresholdChanged) {
                            // If the threshold changed ...
                            if (DeviceModel.parameters.ForwardPower.HoldTimerActive == true) {
                                // And the threshold has returned to the same place ... reset the time
                                LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [THRESHOLD RETURNED - QualifierValid=" + QualifierValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                                ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerHoldState", { "ForwardPowerHoldState": "OFF" });
                                DeviceModel.parameters.ForwardPower.HoldTimerActive = false;
                            }
                        }
                    }
                }
                //#endregion HoldTime Logic for Qualifier USED & TRUE
            }
            else {
                //#region HoldTime Logic for Qualifer USED & FALSE
                QualifierValid = false;
                LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");

                if (DeviceModel.parameters.ForwardPower.HoldTime > 0) {
                    // If there is a hold time ...
                    if (DeviceModel.parameters.ForwardPower.HoldTimerActive === true) {
                        // Reset the hold timer ... it will have to be requalified
                        DeviceModel.parameters.ForwardPower.HoldTimerActive = false;
                        LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [THRESHOLD RETURNED - QualifierValid=" + QualifierValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerHoldState", { "ForwardPowerHoldState": "OFF" });
                    }
                }
                //#endregion HoldTime Logic for Qualifer USED & FALSE
            }
            //#endregion HoldTime Logic for Qualifier USED
            //#endregion Qualifier USED logic
        }
        else {
            // Qualifier NOT USED logic
            //#region Qualifier NOT USED logic
            // HoldTime Logic for Qualifier NOT USED
            //#region HoldTime Logic for QUALIFIER NOT USED
            QualifierValid = true;
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: Qualifier=NOT_USED]");
            if (DeviceModel.parameters.ForwardPower.HoldTime > 0) {
                // If there is a hold time ...
                if (DeviceModel.parameters.ForwardPower.HoldTimerActive === false) {
                    // If the hold timer was not previously running ...
                    if (ThresholdChanged) {
                        // And the threshold changed ... start the timer
                        DeviceModel.parameters.ForwardPower.HoldTimer = timeChanged;
                        DeviceModel.parameters.ForwardPower.HoldTimerActive = true;
                        LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerHoldState", { "ForwardPowerHoldState": "HOLDING" });
                    }
                }
                else {
                    //  And the threshold has returned to same place ... reset the timer
                    // If the hold timer was previously running ...
                    if (!ThresholdChanged) {
                        LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [THRESHOLD RETURNED - QualifierValid=" + QualifierValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerHoldState", { "ForwardPowerHoldState": "OFF" });
                        DeviceModel.parameters.ForwardPower.HoldTimerActive = false;
                    }
                }
            }
            //#endregion HoldTime Logic for QUALIFIER NOT USED
            //#endregion Qualifier NOT USED logic
        }
        //#endregion Qualification logic

        // HoldTime logic
        //  - HoldTimeValid = true or false, depending on:
        //      - true: if holdtimer is used & >= holdtime
        //      - true: if holdtimer is not used
        //      - false: all other conditions
        //#region HoldTime logic
        var HoldTimeValid = false;
        var HoldTimeActive = false;

        if (DeviceModel.parameters.ForwardPower.HoldTime > 0) {
            // Using a hold time ...
            //#region HoldTime USED
            if (DeviceModel.parameters.ForwardPower.HoldTimerActive == true) {
                // Hold time is valid ...
                HoldTimeActive = true;
                HoldTimeChanged = (timeChanged - DeviceModel.parameters.ForwardPower.HoldTimer) / 1000;
                if (HoldTimeChanged > DeviceModel.parameters.ForwardPower.HoldTime) {
                    // HoldTime reached - briefly turn on HoldTime button, then off again
                    HoldTimeValid = true;
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [HOLDTIME LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerHoldState", { "ForwardPowerHoldState": "ON" });
                    // Turn on, then turn off HoldTime button
                    setTimeout(function () {
                        ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerHoldState", { "ForwardPowerHoldState": "OFF" });
                    }, 1000);

                } else {
                    // HoldTime not reached
                    LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [HOLDTIME LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerHoldState", { "ForwardPowerHoldState": "HOLDING" });
                }
            }
            //#endregion HoldTime USED
        } else {
            // Not using a hold time ...
            //#region HoldTime NOT USED
            HoleTimeActive = false;
            HoldTimeValid = true;
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [HOLDTIME LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
            //#endregion HoldTime NOT USED
        }
        //#endregion HoldTime logic

        // Publish logic (real-time info and SNMP traps)
        //  Refer to the 'Crest - Real-time updates and SNMP traps.xlsx' document for cases
        //#region Publish logic
        var FormulaIndex;
        if (QualifierValid && HoldTimeValid) {
            // If the change is qualified and holdtime is valid ...
            //#region QualifierValid=TRUE and HoldTimeValid=TRUE preamble

            // Update the raw and associated formula values because all filtering is valid ...
            DeviceModel.parameters.ForwardPower.RawValue = readValue;
            //#region Update the formula values
            for (i = 0; i < DeviceModel.parameters.ForwardPower.Formulas.length; i++) {
                _expression = mathparser.parse(DeviceModel.parameters.ForwardPower.Formulas[i].Formula);
                DeviceModel.parameters.ForwardPower.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.ForwardPower.RawValue })).toFixed(DeviceModel.parameters.ForwardPower.Formulas[i].Decimals);
            }
            //#endregion Update the formula values
            //#endregion QualifierValid=TRUE and HoldTimeValid=TRUE preamble

            if (ThresholdChanged) {
                // Threshold change is qualified and hold time is valid, so record the changes
                //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE

                // Update Threshold
                DeviceModel.parameters.ForwardPower.TimeChanged = timeChanged;
                DeviceModel.parameters.ForwardPower.ThresholdState = NewThreshold;
                DeviceModel.parameters.ForwardPower.HoldTimerActive = false;// If the threshold changed, set HoldTimerState back to false to restart HoldTimer

                // Update real-time value, including SNMP because of threshold change
                //  If this is the first time through, only report out of normal values
                //  All other times report any changes
                //if ((FORWARD_POWER_FIRST_READ == false) ||
                //    (FORWARD_POWER_FIRST_READ == true && DeviceModel.parameters.ForwardPower.ThresholdState != 0)) {
                ThisDeviceObject.sendNotificationEvents(DeviceModel.port, DeviceModel.uuid, DeviceModel.parameters.ForwardPower.SendEvents, "ForwardPower", { "ForwardPower": DeviceModel.parameters.ForwardPower });
                //}

                // Persist the new settings
                ThisDeviceObject.saveDeviceModel(DeviceModel);

                // Log the changes
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [THRESHOLD CHANGE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [DATABASE UPDATE] =" + JSON.stringify(DeviceModel.parameters.ForwardPower));

                //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
            }
            else {
                // No threshold changed, so just send real time update
                //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE

                // Update real-time value, if there was a value change or we're holding
                if (PreviousRawValue != DeviceModel.parameters.ForwardPower.RawValue) {
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPower", { "ForwardPower": DeviceModel.parameters.ForwardPower });
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [REAL-TIME UPDATE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                }
                //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
            }
        } else {
            // If we're qualified and changes are happening within the threshold value, send an update
            if (QualifierValid) {
                // No threshold changed, so just send real time update
                //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                //#region QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
                // Update the raw and associated formula values because all filtering is valid ...
                DeviceModel.parameters.ForwardPower.RawValue = readValue;
                //#region Update the formula values
                for (i = 0; i < DeviceModel.parameters.ForwardPower.Formulas.length; i++) {
                    _expression = mathparser.parse(DeviceModel.parameters.ForwardPower.Formulas[i].Formula);
                    DeviceModel.parameters.ForwardPower.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.ForwardPower.RawValue })).toFixed(DeviceModel.parameters.ForwardPower.Formulas[i].Decimals);
                }
                //#endregion Update the formula values

                // Update real-time value, if there was a value change or we're holding
                if (PreviousRawValue != DeviceModel.parameters.ForwardPower.RawValue) {
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPower", { "ForwardPower": DeviceModel.parameters.ForwardPower });
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [REAL-TIME UPDATE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold == PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ForwardPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ForwardPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ForwardPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ForwardPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                }
                //#endregion QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
            } else {
                // Not qualified, don't do anything ...
                LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateForwardPower: [Qualifier=USED, QualifierValid=FALSE]");
            }
        }
        //#endregion Publish logic
    }

    this.updateReflectedPower = function (readValue) {
        var timeChanged = moment();

        // Threshold Change logic
        //  - ThresholdChanged = true or false, depending if we've changed a threshold
        //#region Threshold Change Logic
        var ThresholdChanged = false;
        var HoldTimeChanged = 0;

        var PreviousRawValue = DeviceModel.parameters.ReflectedPower.RawValue;
        var PreviousThreshold = DeviceModel.parameters.ReflectedPower.ThresholdState;
        var NewThreshold;
        if (readValue < DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaLowThreshold) {
            NewThreshold = DEVICE_BELOW_THRESHOLD;
        }
        else if (readValue > DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaHighThreshold) {
            NewThreshold = DEVICE_ABOVE_THRESHOLD;
        } else {
            NewThreshold = DEVICE_IN_RANGE;
        }
        if (PreviousThreshold != NewThreshold) {
            ThresholdChanged = true;
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: Reflected Power=" + DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaValue + " " + DeviceModel.parameters.ReflectedPower.Formulas[0].Units + " Raw=" + DeviceModel.parameters.ReflectedPower.RawValue);
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: Threshold change - new=" + ThresholdValues[NewThreshold] + " old=" + ThresholdValues[PreviousThreshold]);
        } else {
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: Reflected Power=" + DeviceModel.parameters.ReflectedPower.Formulas[0].FormulaValue + " " + DeviceModel.parameters.ReflectedPower.Formulas[0].Units + " Raw=" + DeviceModel.parameters.ReflectedPower.RawValue);
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: Threshold change - new=" + ThresholdValues[NewThreshold] + " old=" + ThresholdValues[PreviousThreshold]);
        }
        //#endregion Threshold Change Logic

        // Qualification logic
        //  - QualifierValid = true or false, depending if qualified by digital input
        //  - DeviceModel.parameters.ReflectedPower.HoldTimerActive = true or false, depending on qualifications status
        //  - DeviceModel.parameters.ReflectedPower.HoldTimer = holds time of when first qualified, with or without qualifier
        //#region Qualification logic
        var QualifierValid = false;

        if (DeviceModel.parameters.ReflectedPower.Qualifier > 0) {
            // Qualifier USED logic
            //#region Qualifier USED logic

            // HoldTime Logic for Qualifier USED
            //#region HoldTime Logic for Qualifier USED
            if (DeviceModel.parameters.ReflectedPower.QualifierValue > 0) {
                //#region HoldTime Logic for Qualifier USED & TRUE
                QualifierValid = true;
                LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [QUALIFER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");

                if (DeviceModel.parameters.ReflectedPower.HoldTime > 0) {
                    // If there is a hold time ...
                    if (DeviceModel.parameters.ReflectedPower.HoldTimerActive === false) {
                        // If the hold timer not previously running ...
                        if (ThresholdChanged) {
                            // And the threshold changed ... start the timer
                            DeviceModel.parameters.ReflectedPower.HoldTimer = timeChanged;
                            DeviceModel.parameters.ReflectedPower.HoldTimerActive = true;
                            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                            ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerHoldState", { "ReflectedPowerHoldState": "HOLDING" });
                        } else {
                            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                            ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerHoldState", { "ReflectedPowerHoldState": "OFF" });
                        }
                    }
                    else {
                        // If the hold timer was already running ...
                        if (!ThresholdChanged) {
                            // If the threshold changed ...
                            if (DeviceModel.parameters.ReflectedPower.HoldTimerActive == true) {
                                // And the threshold has returned to the same place ... reset the time
                                LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [THRESHOLD RETURNED - QualifierValid=" + QualifierValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                                ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerHoldState", { "ReflectedPowerHoldState": "OFF" });
                                DeviceModel.parameters.ReflectedPower.HoldTimerActive = false;
                            }
                        }
                    }
                }

                //#endregion HoldTime Logic for Qualifier USED & TRUE
            }
            else {
                //#region HoldTime Logic for Qualifer USED & FALSE
                QualifierValid = false;
                LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");

                if (DeviceModel.parameters.ReflectedPower.HoldTime > 0) {
                    // If there is a hold time ...
                    if (DeviceModel.parameters.ReflectedPower.HoldTimerActive === true) {
                        // Reset the hold timer ... it will have to be requalified
                        DeviceModel.parameters.ReflectedPower.HoldTimerActive = false;
                        LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [THRESHOLD RETURNED - QualifierValid=" + QualifierValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerHoldState", { "ReflectedPowerHoldState": "OFF" });
                    }
                }
                //#endregion HoldTime Logic for Qualifer USED & FALSE
            }
            //#endregion HoldTime Logic for Qualifier USED
            //#endregion Qualifier USED logic
        }
        else {
            // Qualifier NOT USED logic
            //#region Qualifier NOT USED logic
            // HoldTime Logic for Qualifier NOT USED
            //#region HoldTime Logic for QUALIFIER NOT USED
            QualifierValid = true;
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: Qualifier=NOT_USED]");
            if (DeviceModel.parameters.ReflectedPower.HoldTime > 0) {
                // If there is a hold time ...
                if (DeviceModel.parameters.ReflectedPower.HoldTimerActive === false) {
                    // If the hold timer was not previously running ...
                    if (ThresholdChanged) {
                        // And the threshold changed ... start the timer
                        DeviceModel.parameters.ReflectedPower.HoldTimer = timeChanged;
                        DeviceModel.parameters.ReflectedPower.HoldTimerActive = true;
                        LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerHoldState", { "ReflectedPowerHoldState": "HOLDING" });
                    }
                }
                else {
                    //  And the threshold has returned to same place ... reset the timer
                    // If the hold timer was previously running ...
                    if (!ThresholdChanged) {
                        LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [THRESHOLD RETURNED - QualifierValid=" + QualifierValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                        ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerHoldState", { "ReflectedPowerHoldState": "OFF" });
                        DeviceModel.parameters.ReflectedPower.HoldTimerActive = false;
                    }
                }
            }
            //#endregion HoldTime Logic for QUALIFIER NOT USED
            //#endregion Qualifier NOT USED logic
        }
        //#endregion Qualification logic

        // HoldTime logic
        //  - HoldTimeValid = true or false, depending on:
        //      - true: if holdtimer is used & >= holdtime
        //      - true: if holdtimer is not used
        //      - false: all other conditions
        //#region HoldTime logic
        var HoldTimeValid = false;
        var HoldTimeActive = false;

        if (DeviceModel.parameters.ReflectedPower.HoldTime > 0) {
            // Using a hold time ...
            //#region HoldTime USED
            if (DeviceModel.parameters.ReflectedPower.HoldTimerActive == true) {
                // Hold time is valid ...
                HoldTimeActive = true;
                HoldTimeChanged = (timeChanged - DeviceModel.parameters.ReflectedPower.HoldTimer) / 1000;
                if (HoldTimeChanged > DeviceModel.parameters.ReflectedPower.HoldTime) {
                    // HoldTime reached - briefly turn on HoldTime button, then off again
                    HoldTimeValid = true;
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [HOLDTIME LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerHoldState", { "ReflectedPowerHoldState": "ON" });
                    // Turn on, then turn off HoldTime button
                    setTimeout(function () {
                        ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerHoldState", { "ReflectedPowerHoldState": "OFF" });
                    }, 1000);

                } else {
                    // HoldTime not reached
                    LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [HOLDTIME LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerHoldState", { "ReflectedPowerHoldState": "HOLDING" });
                }
            }
            //#endregion HoldTime USED
        } else {
            // Not using a hold time ...
            //#region HoldTime NOT USED
            HoleTimeActive = false;
            HoldTimeValid = true;
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [HOLDTIME LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
            //#endregion HoldTime NOT USED
        }
        //#endregion HoldTime logic

        // Publish logic (real-time info and SNMP traps)
        //  Refer to the 'Crest - Real-time updates and SNMP traps.xlsx' document for cases
        //#region Publish logic
        var FormulaIndex;
        if (QualifierValid && HoldTimeValid) {
            // If the change is qualified and holdtime is valid ...
            //#region QualifierValid=TRUE and HoldTimeValid=TRUE preamble

            // Update the raw and associated formula values because all filtering is valid ...
            DeviceModel.parameters.ReflectedPower.RawValue = readValue;
            //#region Update the formula values
            for (i = 0; i < DeviceModel.parameters.ReflectedPower.Formulas.length; i++) {
                _expression = mathparser.parse(DeviceModel.parameters.ReflectedPower.Formulas[i].Formula);
                DeviceModel.parameters.ReflectedPower.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.ReflectedPower.RawValue })).toFixed(DeviceModel.parameters.ReflectedPower.Formulas[i].Decimals);
            }
            //#endregion Update the formula values
            //#endregion QualifierValid=TRUE and HoldTimeValid=TRUE preamble

            if (ThresholdChanged) {
                // Threshold change is qualified and hold time is valid, so record the changes
                //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE

                // Update Threshold
                DeviceModel.parameters.ReflectedPower.TimeChanged = timeChanged;
                DeviceModel.parameters.ReflectedPower.ThresholdState = NewThreshold;
                DeviceModel.parameters.ReflectedPower.HoldTimerActive = false;// If the threshold changed, set HoldTimerState back to false to restart HoldTimer

                // Update real-time value, including SNMP because of threshold change
                //  If this is the first time through, only report out of normal values
                //  All other times report any changes
                //if ((FORWARD_POWER_FIRST_READ == false) ||
                //    (FORWARD_POWER_FIRST_READ == true && DeviceModel.parameters.ReflectedPower.ThresholdState != 0)) {
                ThisDeviceObject.sendNotificationEvents(DeviceModel.port, DeviceModel.uuid, DeviceModel.parameters.ReflectedPower.SendEvents, "ReflectedPower", { "ReflectedPower": DeviceModel.parameters.ReflectedPower });
                //}

                // Persist the new settings
                ThisDeviceObject.saveDeviceModel(DeviceModel);

                // Log the changes
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [THRESHOLD CHANGE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [DATABASE UPDATE] =" + JSON.stringify(DeviceModel.parameters.ReflectedPower));

                //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
            }
            else {
                // No threshold changed, so just send real time update
                //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE

                // Update real-time value, if there was a value change or we're holding
                if (PreviousRawValue != DeviceModel.parameters.ReflectedPower.RawValue) {
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPower", { "ReflectedPower": DeviceModel.parameters.ReflectedPower });
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [REAL-TIME UPDATE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                }
                //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
            }
        } else {
            // If we're qualified and changes are happening within the threshold value, send an update
            if (QualifierValid) {
                // No threshold changed, so just send real time update
                //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                //#region QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
                // Update the raw and associated formula values because all filtering is valid ...
                DeviceModel.parameters.ReflectedPower.RawValue = readValue;
                //#region Update the formula values
                for (i = 0; i < DeviceModel.parameters.ReflectedPower.Formulas.length; i++) {
                    _expression = mathparser.parse(DeviceModel.parameters.ReflectedPower.Formulas[i].Formula);
                    DeviceModel.parameters.ReflectedPower.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.ReflectedPower.RawValue })).toFixed(DeviceModel.parameters.ReflectedPower.Formulas[i].Decimals);
                }
                //#endregion Update the formula values

                // Update real-time value, if there was a value change or we're holding
                if (PreviousRawValue != DeviceModel.parameters.ReflectedPower.RawValue) {
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPower", { "ReflectedPower": DeviceModel.parameters.ReflectedPower });
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [REAL-TIME UPDATE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold == PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.ReflectedPower.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.ReflectedPower.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.ReflectedPower.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.ReflectedPower.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                }
                //#endregion QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
            } else {
                // Not qualified, don't do anything ...
                LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateReflectedPower: [Qualifier=USED, QualifierValid=FALSE]");
            }
        }
        //#endregion Publish logic
    }

    this.updateVSWR = function (readValue) {

        // Checks to ensure that VSWR value is good (Mantis 1021 - as per Bird's instructions on evaluating VSWR)
        //  1. If ReflectedPower > ForwardPower then VSWR=1 (only if we're actually reading ForwardPower and ReflectedPower)
        if (DeviceModel.parameters.ReflectedPower.UnqualifiedValue > DeviceModel.parameters.ForwardPower.UnqualifiedValue) {
            // Don't log if this error was repeated
            if (VSWRStatus != VSWR_ERROR_REFLECTED_GREATER_FORWARD && DeviceModel.parameters.VSWR.RawValue!=1) {
                LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: Adjusted VSWR read value to 1 (ReflectedPower [" + DeviceModel.parameters.ReflectedPower.UnqualifiedValue.toFixed(3) + "W]>ForwardPower [" + DeviceModel.parameters.ForwardPower.UnqualifiedValue.toFixed(3) + "W]), Raw VSWR=" + Number(readValue).toFixed(2) + " Filtered VSWR=1.00");
            }
            readValue = 1
            VSWRStatus = VSWR_ERROR_REFLECTED_GREATER_FORWARD;
        }
        //  2. If FowardPower < 0.5W then VSWR=1 (only if we're actually reading ForwardPower)
        else if (DeviceModel.parameters.ForwardPower.UnqualifiedValue < 0.5) {
            if (VSWRStatus != VSWR_ERROR_FORWARD_TOO_LOW && DeviceModel.parameters.VSWR.RawValue != 1) {
                LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: Adjusted VSWR read value to 1 (ForwardPower [" + DeviceModel.parameters.ForwardPower.UnqualifiedValue.toFixed(3) + "W] < 0.5W), Raw VSWR=" + Number(readValue).toFixed(2) + " Filtered VSWR=1.00");
            }
            readValue = 1
            VSWRStatus = VSWR_ERROR_FORWARD_TOO_LOW;
        }
        //  3. If VSWR=Infinity then ignore this VSWR reading
        else if (!Number.isFinite(readValue)) {
            if (VSWRStatus != VSWR_ERROR_VSWR_INFINITY) {
                LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: Current VSWR value ignored (Infinity), Raw VSWR=" + Number(readValue).toFixed(1) + " Filtered VSWR=" + DeviceModel.parameters.VSWR.RawValue);
            }
            VSWRStatus = VSWR_ERROR_VSWR_INFINITY;
            return;
        }
        //  4. If VSWR<1 then ignore this VSWR reading
        else if (readValue < 1) {
            if (VSWRStatus != VSWR_ERROR_VSWR_LESS_THAN_1) {
                LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: Current VSWR value ignored (Less than 1), Raw VSWR=" + Number(readValue).toFixed(1) + " Filtered VSWR=" + DeviceModel.parameters.VSWR.RawValue);
            }
            VSWRStatus = VSWR_ERROR_VSWR_LESS_THAN_1;
            return;
        } else {
            // VSWR Reading is good, carry on adjusting the value
            VSWRStatus = VSWR_GOOD;
        }

        // Record the time, in case something changed
        var timeChanged = moment();

        // Threshold Change logic
        //  - ThresholdChanged = true or false, depending if we've changed a threshold
        //#region Threshold Change Logic
        var ThresholdChanged = false;
        var HoldTimeChanged = 0;

        var PreviousRawValue = DeviceModel.parameters.VSWR.RawValue;
        var PreviousThreshold = DeviceModel.parameters.VSWR.ThresholdState;
        var NewThreshold;
        if (readValue > DeviceModel.parameters.VSWR.Formulas[0].FormulaHighThreshold) {
            NewThreshold = DEVICE_MAJOR_THRESHOLD;
        }
        else if (readValue > DeviceModel.parameters.VSWR.Formulas[0].FormulaLowThreshold) {
            NewThreshold = DEVICE_MINOR_THRESHOLD;
        } else {
            NewThreshold = DEVICE_IN_RANGE;
        }
        if (PreviousThreshold != NewThreshold) {
            ThresholdChanged = true;
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: VSWR=" + DeviceModel.parameters.VSWR.Formulas[0].FormulaValue + " " + DeviceModel.parameters.VSWR.Formulas[0].Units + " Raw=" + DeviceModel.parameters.VSWR.RawValue);
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: Threshold change - new=" + ThresholdValues[NewThreshold] + " old=" + ThresholdValues[PreviousThreshold]);
        } else {
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: VSWR=" + DeviceModel.parameters.VSWR.Formulas[0].FormulaValue + " " + DeviceModel.parameters.VSWR.Formulas[0].Units + " Raw=" + DeviceModel.parameters.VSWR.RawValue);
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: Threshold change - new=" + ThresholdValues[NewThreshold] + " old=" + ThresholdValues[PreviousThreshold]);
        }
        //#endregion Threshold Change Logic

        // Qualification logic
        //  - QualifierValid = true (since qualifier not used)
        //#region Qualification logic
        var QualifierValid = true;
        //#endregion Qualification logic

        // HoldTime logic
        //  - HoldTimeValid = true (since hold time not used)
        //#region HoldTime logic
        var HoldTimeValid = true;
        //#endregion HoldTime logic

        // Publish logic (real-time info and SNMP traps)
        //  Refer to the 'Crest - Real-time updates and SNMP traps.xlsx' document for cases
        //#region Publish logic
        var FormulaIndex;
        if (QualifierValid && HoldTimeValid) {
            // If the change is qualified and holdtime is valid ...
            //#region QualifierValid=TRUE and HoldTimeValid=TRUE preamble

            // Update the raw and associated formula values because all filtering is valid ...
            DeviceModel.parameters.VSWR.RawValue = readValue;
            DeviceModel.parameters.VSWR.Formulas[0].FormulaValue = Number(DeviceModel.parameters.VSWR.RawValue).toFixed(DeviceModel.parameters.VSWR.Formulas[0].Decimals);
            //#endregion QualifierValid=TRUE and HoldTimeValid=TRUE preamble

            if (ThresholdChanged) {
                // Threshold change is qualified and hold time is valid, so record the changes
                //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE

                // Update Threshold
                DeviceModel.parameters.VSWR.TimeChanged = timeChanged;
                DeviceModel.parameters.VSWR.ThresholdState = NewThreshold;

                // Update real-time value, including SNMP because of threshold change
                //  If this is the first time through, only report out of normal values
                //  All other times report any changes
                //if ((FORWARD_POWER_FIRST_READ == false) ||
                //    (FORWARD_POWER_FIRST_READ == true && DeviceModel.parameters.VSWR.ThresholdState != 0)) {
                ThisDeviceObject.sendNotificationEvents(DeviceModel.port, DeviceModel.uuid, DeviceModel.parameters.VSWR.SendEvents, "VSWR", { "VSWR": DeviceModel.parameters.VSWR });
                //}

                // Persist the new settings
                ThisDeviceObject.saveDeviceModel(DeviceModel);

                // Log the changes
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: [THRESHOLD CHANGE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.VSWR.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.VSWR.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.VSWR.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.VSWR.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: [DATABASE UPDATE] =" + JSON.stringify(DeviceModel.parameters.VSWR));

                //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
            }
            else {
                // No threshold changed, so just send real time update
                //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE

                // Update real-time value, if there was a value change or we're holding
                if (PreviousRawValue != DeviceModel.parameters.VSWR.RawValue) {
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "VSWR", { "VSWR": DeviceModel.parameters.VSWR });
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateVSWR: [REAL-TIME UPDATE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.VSWR.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.VSWR.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.VSWR.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.VSWR.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                }
                //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
            }
        }
        //#endregion Publish logic
    }

    this.updateTemperature = function (readValue) {
        var timeChanged = moment();

        // Threshold Change logic
        //  - ThresholdChanged = true or false, depending if we've changed a threshold
        //#region Threshold Change Logic
        var ThresholdChanged = false;
        var HoldTimeChanged = 0;

        var PreviousRawValue = DeviceModel.parameters.Temperature.RawValue;
        var PreviousThreshold = DeviceModel.parameters.Temperature.ThresholdState;
        var NewThreshold;
        if (readValue < DeviceModel.parameters.Temperature.Formulas[0].FormulaLowThreshold) {
            NewThreshold = DEVICE_BELOW_THRESHOLD;
        }
        else if (readValue > DeviceModel.parameters.Temperature.Formulas[0].FormulaHighThreshold) {
            NewThreshold = DEVICE_ABOVE_THRESHOLD;
        } else {
            NewThreshold = DEVICE_IN_RANGE;
        }
        if (PreviousThreshold != NewThreshold) {
            ThresholdChanged = true;
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: Temperature=" + DeviceModel.parameters.Temperature.Formulas[0].FormulaValue + " " + DeviceModel.parameters.Temperature.Formulas[0].Units + " Raw=" + DeviceModel.parameters.Temperature.RawValue);
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: Threshold change - new=" + ThresholdValues[NewThreshold] + " old=" + ThresholdValues[PreviousThreshold]);
        } else {
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: Temperature=" + DeviceModel.parameters.Temperature.Formulas[0].FormulaValue + " " + DeviceModel.parameters.Temperature.Formulas[0].Units + " Raw=" + DeviceModel.parameters.Temperature.RawValue);
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: Threshold change - new=" + ThresholdValues[NewThreshold] + " old=" + ThresholdValues[PreviousThreshold]);
        }
        //#endregion Threshold Change Logic

        // Qualification logic
        //  - QualifierValid = true (since no qualifier is used)
        //#region Qualification logic

        // Qualifier NOT USED logic
        //#region Qualifier NOT USED logic
        QualifierValid = true;
        if (DeviceModel.parameters.Temperature.HoldTime > 0) {
            // If there is a hold time ...
            if (DeviceModel.parameters.Temperature.HoldTimerActive === false) {
                // If the hold timer was not previously running ...
                if (ThresholdChanged) {
                    // And the threshold changed ... start the timer
                    DeviceModel.parameters.Temperature.HoldTimer = timeChanged;
                    DeviceModel.parameters.Temperature.HoldTimerActive = true;
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "TemperatureHoldState", { "TemperatureHoldState": "HOLDING" });
                } else {
                    LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "TemperatureHoldState", { "TemperatureHoldState": "OFF" });
                }
            }
            else {
                //  And the threshold has returned to same place ... reset the timer
                // If the hold timer was previously running ...
                if (!ThresholdChanged) {
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [QUALIFIER LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [THRESHOLD RETURNED - QualifierValid=" + QualifierValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "TemperatureHoldState", { "TemperatureHoldState": "OFF" });
                    DeviceModel.parameters.Temperature.HoldTimerActive = false;
                }
            }
        }
        //#endregion Qualifier NOT USED logic

        //#endregion Qualification logic

        // HoldTime logic
        //  - HoldTimeValid = true or false, depending on:
        //      - true: if holdtimer is used & >= holdtime
        //      - true: if holdtimer is not used
        //      - false: all other conditions
        //#region HoldTime logic
        var HoldTimeValid = false;
        var HoldTimeActive = false;

        if (DeviceModel.parameters.Temperature.HoldTime > 0) {
            // Using a hold time ...
            //#region HoldTime USED
            if (DeviceModel.parameters.Temperature.HoldTimerActive == true) {
                // Hold time is valid ...
                HoldTimeActive = true;
                HoldTimeChanged = (timeChanged - DeviceModel.parameters.Temperature.HoldTimer) / 1000;
                if (HoldTimeChanged > DeviceModel.parameters.Temperature.HoldTime) {
                    // HoldTime reached - briefly turn on HoldTime button, then off again
                    HoldTimeValid = true;
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [HOLDTIME LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "TemperatureHoldState", { "TemperatureHoldState": "ON" });
                    // Turn on, then turn off HoldTime button
                    setTimeout(function () {
                        ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "TemperatureHoldState", { "TemperatureHoldState": "OFF" });
                    }, 1000);

                } else {
                    // HoldTime not reached
                    LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [HOLDTIME LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "TemperatureHoldState", { "TemperatureHoldState": "HOLDING" });
                }
            }
            //#endregion HoldTime USED
        } else {
            // Not using a hold time ...
            //#region HoldTime NOT USED
            HoleTimeActive = false;
            HoldTimeValid = true;
            LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [HOLDTIME LOGIC - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
            //#endregion HoldTime NOT USED
        }
        //#endregion HoldTime logic

        // Publish logic (real-time info and SNMP traps)
        //  Refer to the 'Crest - Real-time updates and SNMP traps.xlsx' document for cases
        //#region Publish logic
        var FormulaIndex;
        if (QualifierValid && HoldTimeValid) {
            // If the change is qualified and holdtime is valid ...
            //#region QualifierValid=TRUE and HoldTimeValid=TRUE preamble

            // Update the raw and associated formula values because all filtering is valid ...
            DeviceModel.parameters.Temperature.RawValue = readValue;
            //#region Update the formula values
            for (i = 0; i < DeviceModel.parameters.Temperature.Formulas.length; i++) {
                _expression = mathparser.parse(DeviceModel.parameters.Temperature.Formulas[i].Formula);
                DeviceModel.parameters.Temperature.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.Temperature.RawValue })).toFixed(DeviceModel.parameters.Temperature.Formulas[i].Decimals);
            }
            //#endregion Update the formula values
            //#endregion QualifierValid=TRUE and HoldTimeValid=TRUE preamble

            if (ThresholdChanged) {
                // Threshold change is qualified and hold time is valid, so record the changes
                //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
                //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE

                // Update Threshold
                DeviceModel.parameters.Temperature.TimeChanged = timeChanged;
                DeviceModel.parameters.Temperature.ThresholdState = NewThreshold;
                DeviceModel.parameters.Temperature.HoldTimerActive = false;// If the threshold changed, set HoldTimerState back to false to restart HoldTimer

                // Update real-time value, including SNMP because of threshold change
                //  If this is the first time through, only report out of normal values
                //  All other times report any changes
                //if ((FORWARD_POWER_FIRST_READ == false) ||
                //    (FORWARD_POWER_FIRST_READ == true && DeviceModel.parameters.Temperature.ThresholdState != 0)) {
                ThisDeviceObject.sendNotificationEvents(DeviceModel.port, DeviceModel.uuid, DeviceModel.parameters.Temperature.SendEvents, "Temperature", { "Temperature": DeviceModel.parameters.Temperature });
                //}

                // Persist the new settings
                ThisDeviceObject.saveDeviceModel(DeviceModel);

                // Log the changes
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [THRESHOLD CHANGE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [DATABASE UPDATE] =" + JSON.stringify(DeviceModel.parameters.Temperature));

                //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=TRUE
            }
            else {
                // No threshold changed, so just send real time update
                //  *** CASE 1 : QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 2B: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 3A: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //  *** CASE 4D: QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
                //#region QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE

                // Update real-time value, if there was a value change or we're holding
                if (PreviousRawValue != DeviceModel.parameters.Temperature.RawValue) {
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "Temperature", { "Temperature": DeviceModel.parameters.Temperature });
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [REAL-TIME UPDATE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold != PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                }

                //#endregion QualifierValid=TRUE, HoldTimeValid=TRUE, ThresholdChange=FALSE
            }
        } else {
            // If we're qualified and changes are happening within the threshold value, send an update
            if (QualifierValid) {
                // No threshold changed, so just send real time update
                //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                //  *** CASE 2A: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=FALSE
                //  *** CASE 4C: QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=TRUE
                //#region QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
                // Update the raw and associated formula values because all filtering is valid ...
                DeviceModel.parameters.Temperature.RawValue = readValue;
                //#region Update the formula values
                for (i = 0; i < DeviceModel.parameters.Temperature.Formulas.length; i++) {
                    _expression = mathparser.parse(DeviceModel.parameters.Temperature.Formulas[i].Formula);
                    DeviceModel.parameters.Temperature.Formulas[i].FormulaValue = Number(_expression.evaluate({ x: DeviceModel.parameters.Temperature.RawValue })).toFixed(DeviceModel.parameters.Temperature.Formulas[i].Decimals);
                }
                //#endregion Update the formula values

                // Update real-time value, if there was a value change or we're holding
                if (PreviousRawValue != DeviceModel.parameters.Temperature.RawValue) {
                    ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "Temperature", { "Temperature": DeviceModel.parameters.Temperature });
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [REAL-TIME UPDATE - QualifierValid=" + QualifierValid + ", HoldTimeValid=" + HoldTimeValid + ", ThresholdChanged=" + (NewThreshold == PreviousThreshold) + "] RawValue=" + (DeviceModel.parameters.Temperature.RawValue).toFixed(2) + " Threshold=" + (NewThreshold != -1 ? ThresholdValues[NewThreshold] : "UNDEFINED") + " Qualifier=" + (DeviceModel.parameters.Temperature.Qualifier == 0 ? "'NONE" : "'Digital Input " + (DeviceModel.parameters.Temperature.Qualifier)) + "' HoldTime=" + (DeviceModel.parameters.Temperature.HoldTime == 0 ? "0.00" : HoldTimeChanged.toFixed(2)) + " seconds");
                }
                //#endregion QualifierValid=TRUE, HoldTimeValid=FALSE, ValueChange=TRUE, ThresholdChange=DONTCARE
            } else {
                // Not qualified, don't do anything ...
                LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".updateTemperature: [Qualifier=USED, QualifierValid=FALSE]");
            }
        }
        //#endregion Publish logic
    }

    //#endregion Parameter Filtering

    //#region Utility Functions
    // Saves DeviceModel
    this.saveDeviceModel = function (DeviceModel) {
        fs.truncate(PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json", 0, function () {
            fs.writeFile(PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json", JSON.stringify(DeviceModel), function (error) {
                if (error) {
                    LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".saveDeviceFile: write error '" + PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json'" + error);
                }
                cp.exec("sync", function (error, stdout, stderr) {
                    if (error) {
                        LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".saveDeviceFile: sync error '" + PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json'" + error);
                    } else {
                        LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".saveDeviceFile: Device Updated! '" + PERSIST_DEVICE_MODEL_FS_PATH + DeviceModel.port + ".json'");
                    }
                });
            });
        });
    }

    this.updateQualifierValues = function (digitalInputs) {
        if (DeviceModel.parameters.ForwardPower.Qualifier > 0) {
            DeviceModel.parameters.ForwardPower.QualifierValue = digitalInputs[Number(DeviceModel.parameters.ForwardPower.Qualifier) - 1].Value;
            ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ForwardPowerQualifierValue", { "ForwardPowerQualifierValue": DeviceModel.parameters.ForwardPower.QualifierValue });
        }
        if (DeviceModel.parameters.ReflectedPower.Qualifier > 0) {
            DeviceModel.parameters.ReflectedPower.QualifierValue = digitalInputs[Number(DeviceModel.parameters.ReflectedPower.Qualifier) - 1].Value;
            ThisDeviceObject.sendRealTimeEvents(DeviceModel.port, DeviceModel.uuid, "ReflectedPowerQualifierValue", { "ReflectedPowerQualifierValue": DeviceModel.parameters.ReflectedPower.QualifierValue });
        }
    }
    //#endregion Utility Functions

    //#endregion Class Methods (Device Processing - Process I/O)

    //#endregion Class Methods (Device Processing)

    //#region Class Methods (Eventing)

    //#region Class Methods (Eventing - SendEvent)
    // sendEvents - sends socket, SNMP and SMTP messages
    //  SNMP and SMTP only sent if SendEvents flag has been enabled
    this.sendNotificationEvents = function (port, uuid, sendevents, message, parameter) {

        // Update the UI always, regardless of SendEvents setting
        this.sendRealTimeEvents(port, uuid, message, parameter)

        // Only send events if the SendEvents is enabled
        var parameterValues;
        if (sendevents) {
            switch (message) {
                case "ForwardPower":
                    parameterValues = parameter.ForwardPower;
                    break;
                case "ReflectedPower":
                    parameterValues = parameter.ReflectedPower;
                    break;
                case "VSWR":
                    parameterValues = parameter.VSWR;
                    break;
                case "Temperature":
                    parameterValues = parameter.Temperature;
                    break;
            }

            this.sendSMTP(port, uuid, parameterValues);
            this.sendSNMP(port, uuid, parameterValues);
        } else {
            LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendNotificationEvents: Events NOT sent due to parameter SendEvents setting - Parameter=" + message + " SendEvents=" + sendevents);
        }
    }

    this.UpdateSnmpManager = function (snmp) {
        SnmpManager = snmp;
    }

    this.UpdateSummitInfo = function (info) {
        SummitInfo = info;
    }
    //#endregion Class Methods (Eventing - SendEvent)

    //#region Class Methods (Eventing - Real-time Updates)
    this.sendRealTimeEvents = function (port, uuid, message, parameter) {
        //Logger.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendRealTimeEvents: Port=" + port + " Message=" + message + " Parameter=" + JSON.stringify(parameter));
        SocketIoManager.sockets.emit(DeviceModel.id, { "Port": port, "UUID": uuid, "Message": message, "Parameter": parameter });
    }
    //#endregion Class Methods (Eventing - Real-time Updates)

    //#region Class Methods (Eventing - SMTP)
    this.sendSMTP = function (_port, _uuid, parameter) {

        // Send only if device's SendEventsFlag is set
        if (SendEventsFlag === "FALSE") {
            LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSMTP: Events NOT sent due to device SendEvents setting - SendEventsFlag=" + SendEventsFlag);
            return;
        }

        // Set up the SMTP Email message
        var message = [];
        var portNumber = (Number(_port) + 1);
        message += "Timestamp=" + moment(parameter.TimeChanged).toISOString() + "\n";
        message += "Summit MACID=" + SummitInfo.MACID + "\n";
        message += "Summit Name=" + SummitInfo.Name + "\n";
        message += "Serial Port=" + portNumber + "\n";
        message += "Parameter=" + parameter.Id + "\n";
        message += "Status=" + (ThresholdValues[parameter.ThresholdState]) + "\n";
        message += "Raw Value=" + (parameter.RawValue) + "\n";
        message += "Formula Value=" + (parameter.Formulas[parameter.FormulaIndex].FormulaValue) + " " + (parameter.Formulas[parameter.FormulaIndex].Units) + "\n";
        message += "Low Threshold=" + (parameter.Formulas[parameter.FormulaIndex].FormulaLowThreshold) + " " + (parameter.Formulas[parameter.FormulaIndex].Units) + "\n";
        message += "High Threshold=" + (parameter.Formulas[parameter.FormulaIndex].FormulaHighThreshold) + " " + (parameter.Formulas[parameter.FormulaIndex].Units) + "\n";
        message += "Hold Time=" + (parameter.HoldTime==0? "NOT USED":parameter.HoldTime+" seconds") + "\n";
        message += "Qualifier Used=" + (parameter.Qualifier == 0 ? "NOT USED" : "Digital Input #" + (parameter.Qualifier)) + "\n";

        var eventDescription = "Serial Port " + portNumber + " - Parameter '" + parameter.Id + "' Status='" + ThresholdValues[parameter.ThresholdState] + "'";
        LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSMTP: Sending SMTP Event - " + eventDescription);
        SmtpManager.SendSmtp(message, eventDescription, eventDescription, function (result) {
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSMTP: SMTP Email Event Response - " + eventDescription);
        })
    }
    //#endregion Class Methods (Eventing - SMTP)

    //#region Class Methods (Eventing - SNMP)
    this.sendSNMP = function (_port, _uuid, parameter) {

        // Send only if device's SendEventsFlag is set
        if (SendEventsFlag==="FALSE") {
            LogManager.warn(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Events NOT sent due to device SendEvents setting - SendEventsFlag=" + SendEventsFlag);
            return;
        }

        var address = (Number(_port) + 1);

        // Define the notification
        var snmpCommand;
        var snmpHost;
        var snmpNotification = "";
        snmpNotification += "1.3.6.1.4.1.11902.3.5.1000.1000.0." + (address - 1 + 1000) + " ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.1 s '" + moment().toString() + "' ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.2 i " + address + " ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.3 s '" + parameter.Id + "' ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.4 s '" + (Number(parameter.RawValue).toFixed(parameter.Formulas[0].Decimals)).toString() + "' ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.5 i " + parameter.ThresholdState + " ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.6 s '" + Number(parameter.Formulas[0].FormulaLowThreshold).toString() + "' ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.7 s '" + Number(parameter.Formulas[0].FormulaHighThreshold).toString() + "' ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.8 i " + parameter.HoldTime + " ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.10 i " + parameter.Qualifier + " ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.11 i " + parameter.FormulaIndex + " ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.13 s '" + parameter.Formulas[parameter.FormulaIndex].FormulaValue + "' ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.14 s '" + (parameter.Formulas[parameter.FormulaIndex].Units.length == 0 ? " " : parameter.Formulas[parameter.FormulaIndex].Units) + "' ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.15 s '" + parameter.Formulas[parameter.FormulaIndex].Formula + "' ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.20 s '" + SummitInfo.MACID + "' ";
        snmpNotification += "1.3.6.1.4.1.11902.3.5." + (address) + ".0.21 s '" + SummitInfo.Name + "' ";

        // Send the SNMP notification
        for (i = 0; i < SnmpManager.length; i++) {

            // Skip if this host is not enabled
            if (SnmpManager[i].Enabled === "FALSE") continue;

            // Set up the SNMP trap command
            snmpCommand = "snmptrap ";
            if (SnmpManager[i].SnmpVersion == 2) {
                snmpCommand += "-v 2c ";
                snmpCommand += "-c " + SnmpManager[i].Community + " ";
            }
            else if (SnmpManager[i].SnmpVersion == 3) {
                snmpCommand += "-v 3 ";
                snmpCommand += "-e " + SnmpManager[i].Community + " ";
                snmpCommand += "-u " + SnmpManager[i].AuthUser + " ";
                snmpCommand += "-a " + AuthProtocol[SnmpManager[i].AuthProtocol] + " -A " + SnmpManager[i].AuthPassword + " ";
                snmpCommand += "-x " + PrivProtocol[SnmpManager[i].PrivProtocol] + " -X " + SnmpManager[i].PrivPassword + " ";
            } else continue;

            // Send trap to host ...
            snmpHost = SnmpManager[i].IpAddress + " '' ";

            var eventDescription = "Serial Port " + address + " - Parameter '" + parameter.Id + "' Status='" + ThresholdValues[parameter.ThresholdState] + "'";
            LogManager.info(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: Event to '" + SnmpManager[i].HostName + "' event=" + eventDescription);
            LogManager.trace(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP: " + snmpCommand + snmpHost + snmpNotification);

            cp.exec(snmpCommand + snmpHost + snmpNotification, function (error, stdout, stderr) {
                if (error !== null) {
                    LogManager.error(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP. " + error);
                } else {
                    LogManager.debug(DeviceModel.id + "[" + DeviceModel.port + "," + DeviceModel.uuid + "]" + ".sendSNMP. " + stdout);
                }
            });
        }
    }
    //#endregion Class Methods (Eventing - SNMP)

    //#endregion Class Methods (Eventing)

    //#endregion Class Methods   
}

//#region Device Constructor
// Device constructor to ensure unique objects
//  Uses the following signature:
//      port        - port on which device is installed [0 indexed]
//      loadFile    - how to load the model, default or from persisted file
//      logger      - system logger object
//      io          - system socket io object
//      smtpManager - SMTP messaging object
//      netSnmp     - SNMP infrastructure object
module.exports = function (port, loadFile, logger, io, info, smtp, snmp, callback) {
    var DeviceObject = new Device(port, loadFile, logger, io, info, smtp, snmp);
    callback(DeviceObject);
}
//#endregion Device Constructor
//#region Device Class: Variables, Instantiation, Methods and Constructor

// Exported Methods
//#region Exported Methods

// Basic Info
//#region Basic Info
Device.prototype.SetPort = function (port) {
    this.setPort(port);
}

Device.prototype.GetPort = function () {
    return this.getPort()
}

Device.prototype.GetDeviceId = function () {
    return this.getDeviceId()
}

Device.prototype.GetDeviceModel = function () {
    return this.getDeviceModel()
}

Device.prototype.GetDeviceInformation = function (callback) {
    var deviceInfo = this.getDeviceInformation();
    callback(deviceInfo);
}

//#endregion Basic Info

// Enable/Disable/Reset/Zero
//#region Enable/Disable/Reset/Zero
Device.prototype.Enable = function (enabled) {
    var id = this.getDeviceId();
    LogManager.trace(id + ".Enable: EnabledFlag='" + enabled + "'");
    this.enable((enabled=="TRUE"?true:false));
}

Device.prototype.Reset = function (callback) {
    this.reset(function (status) {
        callback(status);
    });
}

Device.prototype.Zero = function (callback) {
    this.zero(function (status) {
        callback(status);
    });
}

//#endregion Enable/Disable/Reset/Zero

// Send Events
//#region Send Events
Device.prototype.SendEvents = function (sendevents) {
    var id = this.getDeviceId();
    LogManager.trace(id + ".SendEvents: SendEventsFlag='" + sendevents + "'");
    this.sendEvents((sendevents=="TRUE"?true:false));
}
//#endregion Send Events

// Update Device
//#region Update Device

// UpdateDeviceModel
Device.prototype.UpdateDeviceModel = function (port, model, callback) {
    var id = this.getDeviceId();
    this.updateDeviceModel(port, model, function (model) {
        LogManager.trace(id + ".UpdateDeviceModel: Updated DeviceModel=" + JSON.stringify(model));
        callback(model);
    });
}

Device.prototype.UpdateQualifierValues = function (digitalInputs) {
    this.updateQualifierValues(digitalInputs);
}

Device.prototype.UpdateSnmpManager = function (snmp) {
    this.updateSnmpManager(snmp);
}

Device.prototype.UpdateSummitInfo = function (info) {
    this.updateSummitInfo(info);
}
//#endregion Update Device

//#endregion Exported Methods